@@ -1,18 +1,19 @@
 import numpy as np
+from numpy.core.shape_base import block
 from .. import ffmpeg, probe, utils, configure
 
 
 class SimpleVideoReader:
     def __init__(self, url, stream_id=0, **options) -> None:
 
-        args = configure.input_timing(url, vstream_id=stream_id, **options)
+        args = configure.input_timing({}, url, vstream_id=stream_id, **options)
 
         args, reader_cfg = configure.video_io(
+            args,
             url,
             stream_id,
             output_url="-",
             format="rawvideo",
-            ffmpeg_args=args,
             excludes=["frame_rate"],
             **options,
         )
@@ -85,6 +86,7 @@ class SimpleVideoWriter:
 
         else:
             args = configure.input_timing(
+                {},
                 "-",
                 vstream_id=0,
                 excludes=("start", "end", "duration"),
@@ -92,9 +94,9 @@ class SimpleVideoWriter:
             )
 
             configure.video_io(
-                utils.array_to_video_input(data, format="rawvideo"),
+                args,
+                utils.array_to_video_input(self.frame_rate, data, format="rawvideo"),
                 output_url=self.url,
-                ffmpeg_args=args,
                 **self.options,
             )
 
@@ -113,14 +115,19 @@ class SimpleVideoWriter:
 
 class SimpleAudioReader:
     def __init__(self, url, stream_id=0, **options) -> None:
-        args = configure.input_timing(url, astream_id=stream_id, **options)
+        args = configure.input_timing({}, url, astream_id=stream_id, **options)
+
+        start, end = configure.get_audio_range(args, stream_id)
+        if start > 0:
+            # if start time is set, remove to read all samples from the beginning
+            del args["inputs"][0][1]["ss"]
 
         args, reader_cfg = configure.audio_io(
+            args,
             url,
             int(stream_id),
             output_url="-",
             format="rawvideo",
-            ffmpeg_args=args,
             **options,
         )
 
@@ -128,9 +135,17 @@ class SimpleAudioReader:
 
         self.proc = ffmpeg.run(args)
         self.stdout = self.proc.stdout
-        self.eof = False
         self.itemsize = np.dtype(self.dtype).itemsize
         self._nbytes = self.channels * self.itemsize
+        self.remaining = end
+        self.eof = end <= 0
+
+        # if starting mid-stream, drop earlier samples
+        while start > 8192:
+            self.read(8192)
+            start -= 8192
+        if start > 0:
+            self.read(start)
 
     def read(self, nsamples=0):
         """read audio samples
@@ -143,19 +158,31 @@ class SimpleAudioReader:
         :return: frame rate and video frame data (dims: time x rows x cols x pix_comps)
         :rtype: (fractions.Fraction, numpy.ndarray)
         """
+
+        if self.eof:
+            return None
+
         nbytes = self._nbytes * nsamples
-        data = self.stdout.read(nbytes if nbytes > 0 else -1)
-        self.eof = nbytes < 1 or len(data) < nbytes
-        return np.frombuffer(data, dtype=self.dtype).reshape((-1, self.channels))
+        data = np.frombuffer(
+            self.stdout.read(nbytes if nbytes > 0 else -1), dtype=self.dtype
+        ).reshape((-1, self.channels))
+        if self.remaining <= data.shape[0]:
+            self.eof = True
+            data = data[: self.remaining, :]
+        self.remaining -= data.shape[0]
+        return data
+
+    def readiter(self, nsamples):
+        while not self.eof:
+            yield self.read(nsamples)
 
     def close(self):
-        if not self.eof:
-            self.proc.terminate()
+        self.proc.terminate()
         self.proc.wait()
 
 
 class SimpleAudioWriter:
-    def __init__(self, url, rate, dtype=None, channels=None, **kwargs) -> None:
+    def __init__(self, url, rate, dtype=None, channels=None, **options) -> None:
         self.proc = None
         self.url = url
         self.samples_written = 0
@@ -163,14 +190,13 @@ class SimpleAudioWriter:
             self.sample_rate = rate
             self.dtype = dtype
             self.channels = channels
-            self.options = kwargs
+            self.options = options
         else:
-            self.open(rate,dtype=dtype,channels=channels,**options)
-
+            self.open(rate, dtype=dtype, channels=channels, **options)
 
     def open(self, rate, data=None, dtype=None, channels=None, **options):
         if self.proc:
-            raise Exception('stream is already open')
+            raise Exception("stream is already open")
 
         if data is None:
             data = np.empty((1, channels), dtype=dtype)
@@ -179,6 +205,7 @@ class SimpleAudioWriter:
         self.dtype = data.dtype
         self.channels = data.shape[1] if data.ndim > 1 else 1
         args = configure.input_timing(
+            {},
             "-",
             astream_id=0,
             excludes=("start", "end", "duration"),
@@ -186,9 +213,9 @@ class SimpleAudioWriter:
         )
 
         configure.audio_io(
-            utils.array_to_audio_input(data, format=True),
+            args,
+            utils.array_to_audio_input(rate, data, format=True),
             output_url=self.url,
-            ffmpeg_args=args,
             **self.options,
         )
         self.proc = ffmpeg.run(args, stdout=None, stdin=ffmpeg.PIPE)
@@ -219,7 +246,7 @@ class SimpleAudioWriter:
         else:
             if data.ndim != 2 and data.ndim != 1:
                 raise Exception("audio data must be 1d or 2d numpy.ndarray")
-            self.open(self.rate, data.dtype, data.channels)
+            self.open(self.sample_rate, data)
 
         self.proc.stdin.write(data.tobytes())
         self.samples_written += data.shape[0]
